from typing import cast, Callable, Dict, List, Optional, TextIO, Tuple, Type
from types import ModuleType

import importlib
import os
import shutil
import time
import warnings

import sh

from ..error_state import HasErrorState
from ...modules.artifact_collector import ArtifactCollector
from ...modules.reporter import Reporter
from ...lib.ci_exception import CriticalCiException, SilentAbortException
from ...lib.gravity import Dependency
from ...lib.utils import make_block, Uninterruptible, convert_to_str
from ...lib import utils
from ..output import HasOutput
from ..structure_handler import HasStructure
from . import base_vcs
from .swarm import Swarm

__all__ = [
    "PerforceMainVcs",
    "PerforcePollVcs",
    "PerforceSubmitVcs",
    "catch_p4exception"
]


class P4ExceptionStub(BaseException):
    warnings: List[str]
    value: str


class P4stub(ModuleType):  # replace with proper stubs
    P4Exception: Type[P4ExceptionStub]
    P4: Callable


P4Exception = P4ExceptionStub


def catch_p4exception(ignore_if=None):
    return utils.catch_exception("P4Exception", ignore_if)


class PerforceVcs(base_vcs.BaseVcs, HasOutput, HasStructure, HasErrorState):
    """
    This class contains global functions for interaction with Perforce
    """

    @staticmethod
    def define_arguments(argument_parser):
        parser = argument_parser.get_or_create_group("Perforce",
                                                     "Please read the details about P4 environment variables "
                                                     "in official Helix manual")

        parser.add_argument("--p4-port", "-p4p", dest="port", help="P4 port (e.g. 'myhost.net:1666')", metavar="P4PORT")
        parser.add_argument("--p4-user", "-p4u", dest="user", help="P4 user name", metavar="P4USER")
        parser.add_argument("--p4-password", "-p4P", dest="password", help="P4 password", metavar="P4PASSWD")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.check_required_option("port", """
            The perforce 'port' is not specified.

            The perforce port defines protocol, host and listening port of the perforce
            server. Please specify perforce port by using '--p4-port' ('-p4p') command line
            parameter or by setting P4PORT environment variable.
            """)
        self.check_required_option("user", """
            The perforce user name is not specified.

            The perforce user name is required to authenticate with perforce server. Please
            specify the perforce user name by using '--p4-user' ('-p4u') command line
            parameter or by setting P4USER environment variable.
            """)
        self.check_required_option("password", """
            The perforce password is not specified.

            The perforce password is required to authenticate with perforce server. Please
            specify the perforce password by using '--p4-password' ('-p4P') command line
            parameter or by setting P4PASSWD environment variable.
            """)

        try:
            p4_module = cast(P4stub, importlib.import_module("P4"))
        except ImportError as e:
            text = "Error: using VCS type 'p4' requires official Helix CLI and Python package 'perforce-p4python' " \
                   "to be installed. Please refer to `Prerequisites` chapter of project documentation for " \
                   "detailed instructions"
            raise ImportError(text) from e

        # By putting P4 object to self, we can use it in this or any derived classes without any further imports
        self.p4 = p4_module.P4()
        global P4Exception
        P4Exception = p4_module.P4Exception

    @make_block("Connecting")
    @catch_p4exception()
    def connect(self):
        self.disconnect()
        self.p4.port = self.settings.port
        self.p4.user = self.settings.user
        self.p4.password = self.settings.password

        self.p4.connect()
        self.append_repo_status("Perforce server: " + self.settings.port + "\n\n")

    @make_block("Disconnecting")
    def disconnect(self) -> None:
        with warnings.catch_warnings(record=True) as w:
            self.p4.disconnect()
            if not w:
                return
            if "Not connected" in str(w[0].message):  # We consider it ok for a session to expire or not be created yet
                return
            text = ""
            for line in w:
                text += "\n" + warnings.formatwarning(str(line.message), line.category, line.filename, line.lineno)
            self.structure.fail_current_block("Unexpected warning(s): " + text)
            raise SilentAbortException()

    def finalize(self):
        with Uninterruptible(self.out.log_exception) as run:
            run(self.disconnect)
            run(super().finalize)


class PerforceSubmitVcs(PerforceVcs, base_vcs.BaseSubmitVcs):
    @staticmethod
    def define_arguments(argument_parser):
        parser = argument_parser.get_or_create_group("Perforce")
        parser.add_argument("--p4-client", "-p4c", dest="client", metavar="P4CLIENT",
                            help="Existing P4 client (workspace) name to use for submitting")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.check_required_option("client", """
            Perforce workspace name is not specified.
            
            This parameter is required for submitting changes to the perforce. It defines
            the name of the existing workspace to use for submit. Please specify the
            workspace name by using '--p4-client' ('-p4c') command-line parameter or
            P4CLIENT environment variable.
            """)

    def p4reconcile(self, *args, **kwargs):
        try:
            return self.p4.run_reconcile(*args, **kwargs)
        except P4Exception as e:
            if not e.warnings:
                raise P4Exception from e
            if "no file(s) to reconcile" not in e.warnings[0]:
                raise P4Exception from e
            return []

    def reconcile_one_path(self, file_path, workspace_root, change_id, edit_only):
        # TODO: cover 'not file_path.startswith("/")' case with tests
        if not file_path.startswith("/"):
            file_path = workspace_root + "/" + file_path
        if not file_path.endswith("/") and os.path.isdir(file_path):
            file_path += "/"
        if file_path.endswith("/"):
            file_path += "..."
        if edit_only:
            reconcile_result = self.p4reconcile("-c", change_id, "-e", convert_to_str(file_path))
            if not reconcile_result:
                self.out.log(
                    f"The file was not edited. Skipping '{os.path.relpath(file_path, workspace_root)}'...")
        else:
            reconcile_result = self.p4reconcile("-c", change_id, convert_to_str(file_path))

        for line in reconcile_result:
            # p4reconcile returns list of dicts AND strings if file is opened in another workspace
            # so we catch TypeError if line is not dict
            try:
                if line["action"] == "add":
                    self.p4.run_reopen("-c", change_id, "-t", "+w", line["depotFile"])
            except TypeError:
                self.out.log(line)

    @catch_p4exception()
    def submit_new_change(self, description, file_list, review=False, edit_only=False):
        self.connect()

        if review:
            raise CriticalCiException("'--create-review' option is not supported for Perforce at the moment")

        if not self.p4.run_clients("-e", self.settings.client):
            raise CriticalCiException("Workspace '" + self.settings.client + "' doesn't exist!")
        self.p4.client = self.settings.client
        client = self.p4.fetch_client(self.settings.client)
        workspace_root = client['Root']

        change = self.p4.fetch_change()
        change["Files"] = []
        change["Description"] = description
        change_id = self.p4.save_change(change)[0].split()[1]
        self.out.log(f"Created empty CL {change_id} for reconciling")

        try:
            for file_path in file_list:
                self.reconcile_one_path(file_path, workspace_root, change_id, edit_only)

            current_cl = self.p4.fetch_change(change_id)
            # If no changes were reconciled, there will be no file records in CL dictionary
            if "Files" not in current_cl:
                self.p4.run_change("-d", change_id)
                self.out.log(f"Deleted empty CL {change_id}")
                return 0

            text = "The changes in following files are going to be submitted:"
            for line in current_cl["Files"]:
                text+= "\n\t" + line
            self.out.log(text)
            self.p4.run_submit(current_cl, "-f", "revertunchanged")
        except Exception as e:
            self.p4.run_revert("-k", "-c", change_id, "//...")
            self.p4.run_change("-d", change_id)
            self.out.log(f"Reverted and deleted CL {change_id}")
            raise CriticalCiException(str(e)) from e

        return change_id


class PerforceWithMappings(PerforceVcs):

    @staticmethod
    def define_arguments(argument_parser):
        parser = argument_parser.get_or_create_group("Perforce")

        parser.add_argument("--p4-project-depot-path", "-p4d", dest="project_depot_path", metavar="P4_PATH",
                            help="Depot path to get sources from (starts with '//', ends with '/...'"
                                 "Only supports one path. Cannot be used with '--p4-mappings' option")

        parser.add_argument("--p4-mappings", "-p4m", dest="mappings", action="append", nargs='+',
                            metavar="P4_MAPPINGS",
                            help="P4 mappings. Cannot be used with '--p4-project-depot-path' option. "
                                 "Use the following format: '//depot/path/... /local/path/...', "
                                 "where the right half is the same as in real P4 mappings, "
                                 "but without client name. Just start from client root with one slash. "
                                 "For more than one add several times or split with ',' character")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not getattr(self.settings, "project_depot_path", None) and not getattr(self.settings, "mappings", None):
            self.error("""
                Both P4_PATH (-p4d) and P4_MAPPINGS (-p4m) are not set.
                
                Universum needs one of these parameters to be set in order to download sources.
                """)

        # Convert old-style depot path into mappings
        if self.settings.project_depot_path:
            if self.settings.mappings:
                self.error("Both 'P4_PATH' and 'P4_MAPPINGS' cannot be processed simultaneously")
            mappings = [self.settings.project_depot_path + " /..."]
        else:
            mappings = self.settings.mappings

        self.mappings = utils.unify_argument_list(mappings)


class PerforceMainVcs(PerforceWithMappings, base_vcs.BaseDownloadVcs):
    swarm_factory = Dependency(Swarm)
    artifacts_factory = Dependency(ArtifactCollector)
    reporter_factory = Dependency(Reporter)

    @staticmethod
    def define_arguments(argument_parser):
        parser = argument_parser.get_or_create_group("Perforce")

        parser.add_argument("--p4-client", "-p4c", dest="client", metavar="P4CLIENT",
                            help="P4 client (workspace) name to be created. "
                                 "Use '--p4-force-clean' option to delete this client while finalizing")

        parser.add_argument("--p4-sync", "-p4h", action="append", nargs='+', dest="sync_cls",
                            metavar="SYNC_CHANGELIST",
                            help="Sync (head) CL(s). Just a number will be interpreted as united CL for "
                                 "all added VCS roots. To add a sync CL for specific depot/workspace location, "
                                 "write location in the same format as in P4_MAPPINGS with '@<CL number>' "
                                 "in the end, e.g. '//DEV/Solution/MyProject/...@1234567'. To specify "
                                 "more than one sync CL for several locations, add '--p4-sync' several times "
                                 "or split them with comma")

        parser.add_argument("--p4-shelve", "-p4s", action="append", nargs='+', dest="shelve_cls",
                            metavar="SHELVE_CHANGELIST",
                            help="List of shelve CLs to be applied, separated by comma. "
                                 "--p4-shelve can be added to the command line several times. "
                                 "Also shelve CLs can be specified via additional environment variables: "
                                 "SHELVE_CHANGELIST_1..5")

        parser.add_argument("--p4-force-clean", action="store_true", dest="force_clean",
                            help="**Revert all vcs within '--p4-client' and delete the workspace.** "
                                 "Mandatory for CI environment, otherwise use with caution")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.check_required_option("client", """
            Perforce workspace name is not specified.
        
            This parameter is required for creating temporary workspace for downloading
            project sources. Please specify the workspace name by using '--p4-client'
            ('-p4c') command-line parameter or P4CLIENT environment variable.
            """)

        self.artifacts = self.artifacts_factory()
        self.reporter = self.reporter_factory()
        # self.swarm is initialized by self.code_review()
        self.swarm = None

        self.client_name = self.settings.client
        self.client_root = self.settings.project_root

        self.sync_cls = []
        self.shelve_cls = []
        self.depots = []
        self.client_view = []
        self.mappings_dict = {}

        self.unshelved_files: List[Dict[str, str]] = []
        self.diff_in_files: List[Tuple[Optional[str], Optional[str], Optional[str]]] = []

    def code_review(self):
        self.swarm = self.swarm_factory(self.settings.user, self.settings.password)
        return self.swarm

    def get_related_cls(self, cl_number):
        cl_list = [cl_number]
        description = self.p4.run_describe(cl_number)[0]
        for entry in description['desc'].splitlines():
            if entry.startswith("[Related change IDs]"):
                cl_list.extend([number.strip() for number in entry.strip("[Related change IDs]").split(",")])

        return cl_list

    def expand_workspace_parameters(self):
        # Create a list of depots for sync
        for mapping in self.mappings:
            splat_mapping = mapping.split(" ")
            self.depots.append({"path": splat_mapping[0]})
            self.client_view.append(splat_mapping[0] + " //" + self.client_name + splat_mapping[-1])

        # Finalize the list of depots for sync: merge and define sync CLs
        self.sync_cls = utils.unify_argument_list(self.settings.sync_cls)
        if len(self.sync_cls) == 1 and self.sync_cls[0].isdigit():
            for depot in self.depots:
                depot["cl"] = self.sync_cls[0]

        else:
            for depot in self.depots:
                depot["cl"] = None

            for entry in self.sync_cls:
                splat_entry = entry.split("@")
                if len(splat_entry) != 2 or not all(splat_entry):
                    text = f"Something went wrong when processing sync CL parameter ('{self.settings.sync_cls}')"
                    raise CriticalCiException(text)
                # Remove identical depot entries, mostly for aesthetic reasons
                for index, depot in enumerate(self.depots):
                    if splat_entry[0] == depot["path"]:
                        self.depots.pop(index)
                self.depots.append({"path": splat_entry[0], "cl": splat_entry[1]})

        # Retrieve list of shelved CLs from Swarm and "classic" environment variables
        cls = []
        if self.swarm:
            swarm_cls = self.get_related_cls(self.swarm.settings.change)
            cls.extend(swarm_cls)
        for x in range(1, 6):
            cls.append(os.getenv("SHELVE_CHANGELIST_" + str(x)))
        self.shelve_cls = sorted(list(set(utils.unify_argument_list(self.settings.shelve_cls, additional_list=cls))))

    def p4report(self, report):
        for line in report:
            if isinstance(line, dict):
                self.out.log(line["depotFile"] + " (" + line["action"] + ")")

    @make_block("Creating workspace")
    @catch_p4exception()
    def create_workspace(self):
        self.expand_workspace_parameters()

        if not all((self.client_name, self.client_root, self.client_view)):
            raise CriticalCiException("Workspace is not created. Some of these parameters are missing: "
                                      "client name, root directory or mappings.")

        if self.settings.force_clean:
            self.clean_workspace()

        if self.p4.run_clients("-e", self.client_name):
            raise CriticalCiException("Workspace '" + self.client_name + "' already exists!")

        client = self.p4.fetch_client(self.client_name)
        client["Root"] = convert_to_str(self.client_root)
        client["View"] = self.client_view
        self.p4.save_client(client)
        self.p4.client = self.client_name
        self.out.log("Workspace '" + self.client_name + "' created/updated.")

        self.append_repo_status("Workspace: " + self.client_name + "\n")
        self.append_repo_status("Workspace root: " + self.client_root + "\n")
        self.append_repo_status("Mappings:\n")
        for line in self.client_view:
            self.append_repo_status("    " + line + "\n")

    @make_block("Downloading")
    @catch_p4exception()
    def sync(self):
        self.sources_need_cleaning = True        # pylint: disable=attribute-defined-outside-init
        self.append_repo_status("Sync CLs:\n")

        for idx, depot in enumerate(self.depots):
            if depot["cl"] is None:
                self.out.log("Getting latest CL number for '" + depot["path"] + "'")
                try:
                    depot["cl"] = self.p4.run_changes("-m", "1", "-s", "submitted", depot["path"])[0]["change"]
                except IndexError as e:
                    text = "Error getting latest CL number for '" + depot["path"] + "'"
                    text += "\nPlease check depot path formatting (e.g. '/...' in the end for directories)"
                    raise CriticalCiException(text) from e
                self.out.log("Latest CL: " + depot["cl"])

            line = depot["path"] + '@' + depot["cl"]
            # Set environment variable for each mapping in order of there definition
            os.environ[f"SYNC_CL_{idx}"] = depot["cl"]

            self.out.log("Downloading " + line)
            try:
                result = self.p4.run_sync("-f", line)
            except P4Exception as e:
                if "not in client view" not in str(e):
                    raise CriticalCiException(str(e)) from e

                text = f"{e}\nPossible reasons of this error:"
                text += "\n * Wrong formatting (e.g. no '/...' in the end of directory path)"
                text += "\n * Location in 'SYNC_CHANGELIST' is not actually located inside any of 'P4_MAPPINGS'"
                raise CriticalCiException(text) from e

            self.append_repo_status(f"    {line}\n")
            self.out.log(f"Downloaded {result[0]['totalFileCount']} files.")

    def p4unshelve(self, *args, **kwargs):
        try:
            result = self.p4.run_unshelve(*args, **kwargs)
        except P4Exception as e:
            if "already committed" in str(e) and self.swarm and len(self.shelve_cls) == 1:
                self.out.log("CL already committed")
                self.out.report_build_status("CL already committed")
                self.swarm = None
                raise SilentAbortException(application_exit_code=0) from e  # This scenario does not fail build
            raise
        return result

    @make_block("Unshelving")
    @catch_p4exception()
    def unshelve(self):
        if self.shelve_cls:
            self.append_repo_status("Shelve CLs:")
            for cl in self.shelve_cls:
                self.out.log("Unshelve CL " + cl)
                report = self.p4unshelve("-s", cl, "-f")
                self.map_local_path_to_depot(report)
                self.p4report(report)
                self.append_repo_status(" " + cl)
            self.append_repo_status("\n")

    @catch_p4exception(ignore_if="file(s) up-to-date")
    def check_diff_for_depot(self, depot: str) -> str:
        try:
            p4cmd = sh.Command("p4")
            diff_result = p4cmd("-c", self.settings.client, "-u", self.settings.user,
                                "-P", self.settings.password, "-p", self.settings.port,
                                "diff", depot)
            result: str = utils.trim_and_convert_to_unicode(diff_result.stdout)
        except sh.ErrorReturnCode as e:
            for line in e.stderr.splitlines():
                if not (line.startswith(b"Librarian checkout")
                        or line.startswith(b"Error opening librarian file")
                        or line.startswith(b"Transfer of librarian file")
                        or line.endswith(b".gz: No such file or directory")):
                    raise CriticalCiException(utils.trim_and_convert_to_unicode(e.stderr)) from e
            result = utils.trim_and_convert_to_unicode(e.stdout)
        return result

    def calculate_file_diff(self) -> Optional[List[Dict[str, str]]]:
        action_list: Dict[str, str] = {}
        for timeout in [0, 5, 10, 20, 40, 80]:
            time.sleep(timeout)
            try:
                opened_files = self.p4.run_opened()
                break
            except P4Exception as e:
                if f"Client '{self.client_name}' unknown" not in e.value:
                    raise
                self.out.log(f"Getting file diff via 'p4 opened' failed after {timeout} seconds timeout")
        else:
            self.out.log_stderr("Calculating file diff failed, leaving blank")
            return None

        for entry in opened_files:
            action_list[entry["depotFile"]] = entry["action"]
        if not action_list:
            return [{}]

        result: List[Dict[str, str]] = []
        # Both 'p4 opened' and 'p4 where' entries have same key 'depotFile'
        for entry in self.p4.run_where(list(action_list.keys())):
            result.append({"action": action_list[entry["depotFile"]],
                           "repo_path": entry["depotFile"],
                           "local_path": entry["path"]})
        return result

    @make_block("Checking diff")
    def diff(self) -> None:
        rep_diff: List[str] = []
        for depot in self.depots:
            line: str = depot["path"] + '@' + depot["cl"]
            result: str = self.check_diff_for_depot(line)
            if result:
                rep_diff.append(result + "\n")

        if rep_diff:
            file_name = "REPOSITORY_DIFFERENCE.txt"
            self.append_repo_status("See '" + file_name + "' for details on unshelved changes\n")

            f: TextIO = self.artifacts.create_text_file(file_name)
            for result in rep_diff:
                f.write(result)
            f.close()

    def map_local_path_to_depot(self, report):
        for line in report:
            if isinstance(line, dict):
                abs_path = self.p4.run("where", line["depotFile"])[0]["path"]
                self.mappings_dict[abs_path] = line["depotFile"]

    @make_block("Revert workspace to depot state")
    @catch_p4exception()
    def copy_cl_files_and_revert(self) -> List[Tuple[Optional[str], Optional[str], Optional[str]]]:
        self.unshelved_files = self.p4.run_opened()
        unshelved_path: List[Tuple[Optional[str], Optional[str], Optional[str]]] = []

        unshelved_filtered: List[Dict[str, str]] =\
            [item for item in self.unshelved_files if item["action"] != "move/delete"]

        for item in unshelved_filtered:
            relative: Optional[str] = None
            copied: Optional[str] = None
            absolute: Optional[str]
            if item["action"] == "delete":
                absolute = os.path.join(self.client_root, item["clientFile"].replace("//" + item["client"] + "/", ""))
            else:
                relative = item["clientFile"].replace("//" + item["client"] + "/", "")
                copied = os.path.join(self.client_root, "new_temp", relative)
                absolute = os.path.join(self.client_root, relative)
                try:
                    shutil.copy(absolute, copied)
                except IOError:
                    os.makedirs(os.path.dirname(copied))
                    shutil.copy(absolute, copied)
                # absolute = None to make sure content of 'add' and 'branch' won't participate in diff after revert
                # for 'branch' diff we will assume it is a new file
                # be careful, file for 'add' will be present in repo after revert
                if item["action"] in ["add", "branch"]:
                    absolute = None
            unshelved_path.append((relative, copied, absolute))

        if self.shelve_cls:
            self.p4.run_revert("//...")

            for item, path in zip(unshelved_filtered, unshelved_path):
                relative, copied, absolute = path
                if item["action"] == "move/add":
                    for local, depot in self.mappings_dict.items():
                        if depot == item["movedFile"]:
                            absolute = local
                self.diff_in_files.append((relative, copied, absolute))
        return self.diff_in_files

    def prepare_repository(self):
        self.connect()
        self.create_workspace()
        self.sync()
        self.unshelve()
        self.diff()
        if self.swarm:
            self.swarm.client_root = self.client_root
            self.swarm.mappings_dict = self.mappings_dict

    @make_block("Cleaning workspace", pass_errors=False)
    def clean_workspace(self):
        try:
            self.p4.client = self.client_name
            try:
                report = self.p4.run_revert("-k", "-c", "default", "//...")
                self.p4report(report)
            except P4Exception as e:
                if "file(s) not opened on this client" in e.value:
                    self.out.log("Default CL is empty")
                else:
                    raise
            shelves = self.p4.run_changes("-c", self.client_name, "-s", "shelved")
            for item in shelves:
                self.out.log("Deleting shelve from CL " + item["change"])
                self.p4.run_shelve("-d", "-c", item["change"])
            all_cls = self.p4.run_changes("-c", self.client_name, "-s", "pending")
            for item in all_cls:
                self.out.log("Deleting CL " + item["change"])
                try:
                    self.p4.run_revert("-k", "-c", item["change"], "//...")
                except P4Exception as e:
                    if "file(s) not opened on this client" in e.value:
                        self.out.log(f"CL {item['change']} is empty")
                    else:
                        raise
                self.p4.run_change("-d", item["change"])
            self.p4.delete_client(self.client_name)
            self.out.log(f"Client '{self.client_name}' deleted")
        except P4Exception as e:
            if f"Client '{self.client_name}' unknown" in e.value:
                self.out.log("No client to delete")
            else:
                self.structure.fail_current_block(e.value)

    def finalize(self):
        with Uninterruptible(self.out.log_exception) as run:
            if self.settings.force_clean:
                run(self.connect)
                run(self.clean_workspace)
            run(self.disconnect)
            run(super().finalize)


class PerforcePollVcs(PerforceWithMappings, base_vcs.BasePollVcs):
    def get_changes(self, changes_reference=None, max_number='1'):
        self.connect()

        if not changes_reference:
            changes_reference = {}
        result = {}

        for depot in self.mappings:
            depot_path = depot.split(" ")[0]
            if depot_path not in result:
                result[depot_path] = []

            changes = self.p4.run_changes("-s", "submitted", "-m1", depot_path)
            last_cl = changes[0]["change"]
            reference_cl = changes_reference.get(depot_path, last_cl)

            rev_range_string = depot_path + "@" + reference_cl + ",#head"
            submitted_cls = self.p4.run_changes("-s", "submitted", "-m" + str(max_number), rev_range_string)

            submitted_cls.reverse()
            for cl in submitted_cls:
                result[depot_path].append(cl["change"])

        return result



import copy

from typing import Callable, ClassVar, List, Optional, TypeVar
from typing_extensions import TypedDict
from ..configuration_support import Step, Configuration
from ..lib.ci_exception import SilentAbortException, StepException, CriticalCiException
from ..lib.gravity import Dependency, Module
from .output import HasOutput

__all__ = [
    "HasStructure"
]


class Block:
    """
    >>> b1 = Block('Build for all supported platforms')
    >>> str(b1)
    ' Build for all supported platforms - Success'

    >>> b2 = Block('Build Android', b1)
    >>> str(b2)
    '1. Build Android - Success'

    >>> b3 = Block('Build Tizen', b1)
    >>> str(b3)
    '2. Build Tizen - Success'

    >>> b3.status = 'Failed'
    >>> str(b3)
    '2. Build Tizen - Failed'
    >>> b3.is_successful()
    False

    >>> b4 = Block('Run tests', b2)
    >>> str(b4)
    '1.1. Run tests - Success'
    >>> b4.is_successful()
    True
    >>> b2 is b4.parent
    True
    """

    def __init__(self, name: str, parent: Optional["Block"] = None) -> None:
        self.name: str = name
        self.status: str = "Success"
        self.children: List[Block] = []

        self.parent: Optional[Block] = parent
        self.number: str = ''
        if parent:
            parent.children.append(self)
            self.number = f"{parent.number}{len(parent.children)}."

    def __str__(self) -> str:
        result = self.number + ' ' + self.name
        return f"{result} - {self.status}" if not self.children else result

    def is_successful(self) -> bool:
        return self.status == "Success"


class BackgroundStepInfo(TypedDict):
    name: str
    finalizer: Callable[[], None]
    is_critical: bool


class StructureHandler(HasOutput):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.current_block: Optional[Block] = Block("Universum")
        self.configs_current_number: int = 0
        self.configs_total_count: int = 0
        self.active_background_steps: List[BackgroundStepInfo] = []

    def open_block(self, name: str) -> None:
        new_block = Block(name, self.current_block)
        self.current_block = new_block

        self.out.open_block(new_block.number, name)

    def close_block(self) -> None:
        if self.current_block is not None:
            block: Block = self.current_block
            self.current_block = self.current_block.parent
            self.out.close_block(block.number, block.name, block.status)

    def report_critical_block_failure(self) -> None:
        self.out.report_skipped("Critical step failed. All further configurations will be skipped")

    def report_skipped_block(self, name):
        new_skipped_block = Block(name, self.current_block)
        new_skipped_block.status = "Skipped"

        self.out.report_skipped(new_skipped_block.number + " " + name +
                                " skipped because of critical step failure")

    def fail_current_block(self, error: str = ""):
        block: Block = self.get_current_block()
        self.fail_block(block, error)

    def fail_block(self, block, error: str = ""):
        if error:
            self.out.log_exception(error)
        block.status = "Failed"
        self.out.report_build_problem(block.name + " " + block.status)

    def get_current_block(self):
        return self.current_block

    # The exact block will be reported as failed only if pass_errors is False
    # Otherwise the exception will be passed to the higher level function and handled there
    T = TypeVar('T')

    def run_in_block(self, operation: Callable[..., T],
                     block_name: str, pass_errors: bool, *args, **kwargs) -> Optional[T]:
        result = None
        self.open_block(block_name)
        try:
            result = operation(*args, **kwargs)
        except (SilentAbortException, StepException):
            raise
        except CriticalCiException as e:
            self.fail_current_block(str(e))
            raise SilentAbortException() from e
        except Exception as e:
            if pass_errors is True:
                raise
            self.fail_current_block(str(e))
        finally:
            self.close_block()
        return result

    def execute_one_step(self, configuration: Step,
                         step_executor: Callable, is_critical: bool) -> None:
        # step_executor is [[Step], Step], but referring to Step creates circular dependency
        process = step_executor(configuration)

        process.start()
        if not configuration.background:
            process.finalize()
            return

        self.out.log("This step is marked to be executed in background")
        self.active_background_steps.append({'name': configuration.name,
                                             'finalizer': process.finalize,
                                             'is_critical': is_critical})

    def finalize_background_step(self, background_step: BackgroundStepInfo):
        try:
            background_step['finalizer']()
            self.out.log("This background step finished successfully")
        except StepException:
            if background_step['is_critical']:
                self.out.log_stderr("This background step failed, and as it was critical, "
                                    "all further steps will be skipped")
                return False
            self.out.log_stderr("This background step failed")
        return True

    def execute_steps_recursively(self, parent: Optional[Step], cfg: Configuration,
                                  step_executor: Callable,
                                  skipped: bool = False) -> None:
        # step_executor is [[Step], Step], but referring to Step creates circular dependency
        if parent is None:
            parent = Step()

        step_num_len = len(str(self.configs_total_count))
        child_step_failed = False
        for obj_a in cfg.configs:
            try:
                item: Step = parent + copy.deepcopy(obj_a)

                if obj_a.children:
                    # Here pass_errors=True, because any exception outside executing build step
                    # is not step-related and should stop script executing

                    numbering = f" [ {'':{step_num_len}}+{'':{step_num_len}} ] "
                    step_name = numbering + item.name
                    self.run_in_block(self.execute_steps_recursively, step_name, True,
                                      item, obj_a.children, step_executor, skipped)
                else:
                    self.configs_current_number += 1
                    numbering = f" [ {self.configs_current_number:>{step_num_len}}/{self.configs_total_count} ] "
                    step_name = numbering + item.name
                    if skipped:
                        self.report_skipped_block(step_name)
                        continue

                    if item.finish_background and self.active_background_steps:
                        self.out.log("All ongoing background steps should be finished before next step execution")
                        if not self.report_background_steps():
                            self.report_skipped_block(step_name)
                            skipped = True
                            raise StepException()

                    # Here pass_errors=False, because any exception while executing build step
                    # can be step-related and may not affect other steps
                    self.run_in_block(self.execute_one_step, step_name, False,
                                      item, step_executor, obj_a.critical)
            except StepException:
                child_step_failed = True
                if obj_a.critical:
                    self.report_critical_block_failure()
                    skipped = True
        if child_step_failed:
            raise StepException()

    def report_background_steps(self) -> bool:
        result = True
        for item in self.active_background_steps:
            if not self.run_in_block(self.finalize_background_step,
                                     "Waiting for background step '" + item['name'] + "' to finish...",
                                     True, item):
                result = False
        self.out.log("All ongoing background steps completed")
        self.active_background_steps = []
        return result

    def execute_step_structure(self, configs: Configuration, step_executor) -> None:
        self.configs_total_count = sum(1 for _ in configs.all())

        try:
            self.execute_steps_recursively(None, configs, step_executor)
        except StepException:
            pass

        if self.active_background_steps:
            self.run_in_block(self.report_background_steps, "Reporting background steps", False)


class HasStructure(Module):
    structure_factory: ClassVar = Dependency(StructureHandler)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.structure: StructureHandler = self.structure_factory()



import sys

from .modules.api_support import ApiSupport
from .modules.output.output import MinimalOut
from .lib.gravity import Module, Dependency
from .lib.module_arguments import ModuleArgumentParser

__all__ = ["Api"]


class Api(Module):
    description: str = "Universum API"
    api_support_factory = Dependency(ApiSupport)

    @staticmethod
    def define_arguments(parser: ModuleArgumentParser) -> None:
        parser.add_argument('action', choices=["get-shelves", "file-diff", "swarm"],
                            help="Input some description")

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        try:
            self.api_support: ApiSupport = self.api_support_factory(api_mode=True)
        except EnvironmentError as error:
            sys.stderr.write(str(error) + '\n')
            sys.exit(2)

        self.out = MinimalOut(silent=True)

    def execute(self) -> None:
        if self.settings.action == "file-diff":
            print(self.api_support.get_file_diff())
        else:
            raise NotImplementedError()

    def finalize(self) -> None:
        pass



# pylint: disable = protected-access

import argparse
import os
import sys


class ModuleNamespace(argparse.Namespace):

    def __setattr__(self, name, value):
        if '.' in name:
            group, name = name.split('.', 1)
            ns = getattr(self, group, ModuleNamespace())
            setattr(ns, name, value)
            self.__dict__[group] = ns
        else:
            self.__dict__[name] = value

    def __getattr__(self, name):
        if not '.' in name:
            raise AttributeError("No attribute '" + name + "' in module arguments")

        group, name = name.split('.', 1)
        try:
            ns = self.__dict__[group]
        except KeyError as error:
            raise AttributeError("No attribute '" + name + "' in module arguments") from error
        return getattr(ns, name)


# noinspection PyProtectedMember
class ModuleArgumentGroup(argparse._ArgumentGroup):
    def __init__(self, container, *args, **kwargs):
        self.container = container
        super().__init__(container, *args, **kwargs)

    def _add_action(self, action):
        if action.metavar is not None:
            try:
                if 'sphinx' in sys.modules:
                    action.help += f"\n\nEnvironment variable: ${action.metavar}"
                else:
                    if not isinstance(action, argparse._SubParsersAction):
                        action.help += f" [env: {action.metavar}]"
            except TypeError:
                action.help = f"Also available as [env: {action.metavar}]"
            default = os.environ.get(action.metavar, action.default)

            if isinstance(action, (argparse._AppendAction, argparse._AppendConstAction)):
                if default is not None:
                    action.default = [default]
            else:
                action.default = default

        self.container.prepend_dest(action)

        return super()._add_action(action)

    def add_hidden_argument(self, *args, **kwargs):
        if kwargs.get("is_hidden", True):
            kwargs["help"] = argparse.SUPPRESS

        del kwargs["is_hidden"]
        self.add_argument(*args, **kwargs)

    def prepend_dest(self, action):
        self.container.prepend_dest(action)


# noinspection PyProtectedMember
class ModuleArgumentParser(argparse.ArgumentParser):

    def __init__(self, **kwargs):
        self.groups = {}
        self.dest_prefix = ''
        argparse.ArgumentParser.__init__(self, **kwargs)

    def add_argument_group(self, *args, **kwargs):
        group = ModuleArgumentGroup(self, *args, **kwargs)
        self._action_groups.append(group)
        return group

    def get_or_create_group(self, title=None, description=None):

        if title not in self.groups:
            self.groups[title] = self.add_argument_group(title, description)

        return self.groups[title]

    def parse_args(self, args=None, namespace=None):
        """
        Parse passed arguments based on previous setup
        :param args: argument list to parse
        :param namespace: the resulting object to store settings to
        :return: object populated with all settings
        """
        if namespace is None:
            namespace = ModuleNamespace()

        if args is None:
            args = sys.argv[1:]

        self._add_default_parser(args)

        return argparse.ArgumentParser.parse_args(self, args, namespace)

    def add_hidden_argument(self, *args, **kwargs):
        kwargs["help"] = argparse.SUPPRESS
        self.add_argument(*args, **kwargs)

    def prepend_dest(self, action):
        action.dest = self.dest_prefix + action.dest

    def _get_subparsers_action(self):
        if not self._subparsers:
            return None

        for action in self._subparsers._actions:
            if isinstance(action, argparse._SubParsersAction):
                return action

        return None

    @staticmethod
    def _needs_default_parser(subparsers_action, args):
        try:
            possible_subcommand = args[0]
        except IndexError:
            return True

        for sp_name in subparsers_action._name_parser_map.keys():
            if sp_name == possible_subcommand:
                return False

        return True

    def _add_default_parser(self, args):
        subparsers_action = self._get_subparsers_action()
        if not subparsers_action:
            return
        if not self._needs_default_parser(subparsers_action, args):
            return

        default = "default"
        subparsers_action.add_parser(default)
        args.insert(len(args), default)


class IncorrectParameterError(ValueError):
    pass



import requests
from requests import Response

from ..error_state import HasErrorState
from ...lib import utils
from ...lib.module_arguments import ModuleArgumentParser
from .base_server import BaseServerForHostingBuild, BaseServerForTrigger


__all__ = [
    "TeamcityServer"
]


class TeamcityServer(BaseServerForHostingBuild, BaseServerForTrigger, HasErrorState):
    @staticmethod
    def define_arguments(argument_parser: ModuleArgumentParser) -> None:
        parser = argument_parser.get_or_create_group("TeamCity variables",
                                                     "TeamCity-specific parameters")

        parser.add_argument("--tc-server", "-ts", dest="server_url", metavar="TEAMCITY_SERVER",
                            help="TeamCity server URL", type=str)
        parser.add_argument("--tc-build-id", "-tbi", dest="build_id", metavar="BUILD_ID",
                            help="teamcity.build.id", type=str)
        parser.add_argument("--tc-configuration-id", "-tci", dest="configuration_id",
                            metavar="CONFIGURATION_ID", help="system.teamcity.buildType.id", type=str)
        parser.add_argument("--tc-auth-user-id", "-tcu", dest="user_id",
                            metavar="TC_USER", help="system.teamcity.auth.userId", type=str)
        parser.add_argument("--tc-auth-passwd", "-tcp", dest="passwd",
                            metavar="TC_PASSWD", help="system.teamcity.auth.password", type=str)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.check_required_option("server_url", """
            The URL of the TeamCity server is not specified.

            The URL of the TeamCity server is used for constructing links to builds and
            artifacts in code review system, and to adding build tags if requested. Please
            specify the server URL by using '--tc-server' ('-ts') command line parameter or
            by setting TEAMCITY_SERVER environment variable.
            """)
        self.check_required_option("build_id", """
            The id of the build on TeamCity is not specified.

            The id of the build on TeamCity is used for constructing links to builds and
            artifacts in code review system, and to adding build tags if requested. Please
            specify the id of the build by using '--tc-build-id' ('-tbi') command line
            parameter or by setting BUILD_ID environment variable.

            The recommended way to define this parameter is to set BUILD_ID environment
            variable to %teamcity.build.id% in the settings of the top-level project on the
            TeamCity.
            """)
        self.check_required_option("configuration_id", """
            The id of the configuration on TeamCity is not specified.

            The id of the configuration on TeamCity is used for constructing links to
            artifacts in code review system. Please specify the id of the configuration by
            using '--tc-configuration-id' ('-tci') command line parameter or by setting
            CONFIGURATION_ID environment variable.

            The recommended way to define this parameter is to set CONFIGURATION_ID
            environment variable to %system.teamcity.buildType.id% in the settings of the
            top-level project on the TeamCity.
            """)
        self.check_required_option("user_id", """
            The id of the TeamCity user is not specified.

            The TeamCity user id is used for adding build tags. Please specify the  TeamCity
            user id by using '--tc-auth-user-id' ('-tcu') command line parameter or by
            setting TC_USER environment variable.

            The recommended way to define this parameter is to set TC_USER environment
            variable to %system.teamcity.auth.userId% in the settings of the top-level
            project on the TeamCity.

            Please note that the user must have permission to add tags to build.
            """)
        self.check_required_option("passwd", """
            The password of the TeamCity user is not specified.

            The TeamCity user password is used for adding build tags. Please specify the
            TeamCity user password by using '--tc-auth-passwd' ('-tcp') command line
            parameter or by setting TC_PASSWD environment variable.

            The recommended way to define this parameter is to set TC_PASSWD environment
            variable to %system.teamcity.auth.password% in the settings of the top-level
            project on the TeamCity.
            """)

    def report_build_location(self) -> str:
        tc_build_link: str = self.settings.server_url + "/viewLog.html?tab=buildLog&buildId=" + self.settings.build_id
        return "Here is the link to TeamCity build: " + tc_build_link

    def artifact_path(self, local_artifacts_dir: str, item: str) -> str:
        return self.settings.server_url + "/repository/download/" + self.settings.configuration_id + "/" + \
               self.settings.build_id + ":id/" + item

    def add_build_tag(self, tag: str) -> Response:
        return requests.post(f"{self.settings.server_url}/httpAuth/app/rest/builds/id:{self.settings.build_id}/tags",
                             auth=(self.settings.user_id, self.settings.passwd),
                             data=tag, headers={'Content-Type': 'text/plain'}, verify=False)



import json

from .lib.gravity import Dependency
from .lib.utils import make_block
from .modules import automation_server, vcs
from .modules.output import HasOutput
from .modules.structure_handler import HasStructure

__all__ = ["Poll"]


class Poll(HasOutput, HasStructure):
    description = "Polling module of Universum"
    vcs_factory = Dependency(vcs.PollVcs)
    server_factory = Dependency(automation_server.AutomationServerForTrigger)

    @staticmethod
    def define_arguments(parser):
        parser.add_argument('--file', '-f', dest='db_file', metavar="DB_FILE",
                            help='File to store last known CLs', default="p4poll.json")
        parser.add_argument('--num', '-n', dest='max_number', metavar="MAX_NUMBER",
                            help='Maximum number of CLs processed, default is 10', type=int, default=10)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stored_cls = {}
        self.latest_cls = {}
        self.triggered_cls = set()

        self.vcs = self.vcs_factory()
        self.server = self.server_factory()

    def process_single_mapping(self, depot):
        if len(self.latest_cls[depot]) == 1:
            self.stored_cls[depot] = self.latest_cls[depot].pop()
            self.out.log("No changes detected")
            return

        for change in self.latest_cls[depot]:
            if change == self.stored_cls[depot]:
                self.out.log(f"Commit {change} is latest known")
                continue

            if change in self.triggered_cls:
                self.out.log(f"Commit {change} already processed")
                continue

            self.out.log(f"Detected commit {change}, triggering build")
            self.server.trigger_build(change)
            self.triggered_cls.add(change)
            self.stored_cls[depot] = change

    @make_block("Enumerating changes")
    def execute(self):
        try:
            with open(self.settings.db_file, encoding="utf-8") as db_file:
                self.stored_cls = json.load(db_file)
        except IOError as io_error:
            if io_error.errno == 2:
                pass
            else:
                raise

        try:
            self.latest_cls = self.vcs.driver.get_changes(self.stored_cls, self.settings.max_number)
            for depot in self.latest_cls.keys():
                self.structure.run_in_block(self.process_single_mapping, "Processing depot " + depot, True, depot)
        except NotImplementedError:
            self.out.log("Polling is skipped because current VCS doesn't support it")
        finally:
            with open(self.settings.db_file, "w", encoding="utf-8") as db_file:
                json.dump(self.stored_cls, db_file, indent=4, sort_keys=True)

    def finalize(self):
        self.vcs.finalize()



import argparse

from typing import List

from . import utils


def mypy_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Mypy analyzer")
    parser.add_argument("--config-file", dest="config_file", type=str, help="Specify a configuration file.")
    utils.add_python_version_argument(parser)
    return parser


@utils.sys_exit
@utils.analyzer(mypy_argument_parser())
def main(settings: argparse.Namespace) -> List[utils.ReportData]:
    cmd = [f"python{settings.version}", '-m', 'mypy', '--ignore-missing-imports']
    if settings.config_file:
        cmd.append(f'--config-file={settings.config_file}')
    cmd.extend(settings.file_list)
    output, _ = utils.run_for_output(cmd)
    return mypy_output_parser(output)


def mypy_output_parser(output: str) -> List[utils.ReportData]:
    result: List[utils.ReportData] = []
    for raw_line in output.split('\n')[:-2]:  # last line is summary
        data: List[str] = raw_line.split(':', 2)
        if len(data) != 3:
            raise ValueError("Wrong format of output analyzer data: " + raw_line)
        result.append(utils.ReportData(
            symbol="Mypy type annotation issue",
            message=data[2],
            path=data[0],
            line=int(data[1])
        ))
    return result


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter  # see https://github.com/PyCQA/pylint/issues/259



from requests import Response
from ...lib.gravity import Module

__all__ = [
    "BaseServerForTrigger",
    "BaseServerForHostingBuild"
]


class BaseServerForTrigger(Module):
    """
    Abstract base class for API of triggering builds on automation (CI) server
    """

    def trigger_build(self, revision: str) -> None:  # pylint: disable=no-self-use
        raise RuntimeError("Trigger build function is not defined for current driver.")


class BaseServerForHostingBuild(Module):
    """
    Abstract base class for API of hosting CI builds on automation server
    """

    def add_build_tag(self, tag: str) -> Response:  # pylint: disable=no-self-use
        raise RuntimeError("Tag adding function is not defined for current driver.")

    def report_build_location(self) -> str:
        raise NotImplementedError

    def artifact_path(self, local_artifacts_dir: str, item: str) -> str:
        raise NotImplementedError



import json
import sys
import argparse
import glob
import pathlib
import subprocess

from typing import Any, Callable, List, Optional, Tuple
from typing_extensions import TypedDict

from universum.lib.ci_exception import CiException

ReportData = TypedDict('ReportData', {'path': str, 'message': str, 'symbol': str, 'line': int})


class AnalyzerException(CiException):
    def __init__(self, code: int = 2, message: Optional[str] = None):
        self.code: int = code
        self.message: Optional[str] = message


def analyzer(parser: argparse.ArgumentParser):
    """
    Wraps the analyzer specific data and adds common protocol information:
      --files argument and its processing
      --result-file argument and its processing
    This function exists to define analyzer report interface

    :param parser: Definition of analyzer custom arguments
    :return: Wrapped analyzer with common reporting behaviour
    """
    def internal(func: Callable[[argparse.Namespace], List[ReportData]]) -> Callable[[], List[ReportData]]:
        def wrapper() -> List[ReportData]:
            add_files_argument(parser)
            add_result_file_argument(parser)
            settings: argparse.Namespace = parser.parse_args()
            expand_files_argument(settings)
            issues: List[ReportData] = func(settings)
            report_to_file(issues, settings.result_file)
            return issues

        return wrapper

    return internal


def sys_exit(func: Callable[[], Any]) -> Callable[[], None]:
    """
    Execute target function, wrapping any generated exceptions and reporting them to system output
    Exit with code 0, if target function generates no output
    Exit with code 1, if target function generates any outputs, but executes normally
    Exit with code 2 (or custom), if target function fails with an exception

    This decorator is used for analyzer modules to provide normal script interface
    Note: while debugging, remove it from the analyzer code to see the full error state

    :param func: Target function to execute
    :return: None

    >>> def _raise_common() -> None:
    ...     raise Exception()
    >>> def _raise_custom() -> None:
    ...     raise AnalyzerException(code=3)
    >>> def wrap_system_exit(f: Callable) -> int:
    ...     try:
    ...         f()
    ...     except SystemExit as se:
    ...         return se.code
    >>> wrap_system_exit(sys_exit(lambda: None))
    0
    >>> wrap_system_exit(sys_exit(lambda: 'pass'))
    1
    >>> wrap_system_exit(sys_exit(_raise_common))
    2
    >>> wrap_system_exit(sys_exit(_raise_custom))
    3
    """
    def wrapper() -> None:
        exit_code: int
        try:
            res = func()
            exit_code = 1 if res else 0
        except Exception as e:
            message: Optional[str] = getattr(e, 'message', None)
            exit_code = getattr(e, 'code', 2)
            if message:
                sys.stderr.write(message)
            else:
                sys.stderr.write(str(e))
        sys.exit(exit_code)

    return wrapper


def run_for_output(cmd: List[str]) -> Tuple[str, str]:
    result = subprocess.run(cmd, universal_newlines=True,  # pylint: disable=subprocess-run-check
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if result.stderr and not result.stdout:
        raise AnalyzerException(code=result.returncode, message=result.stderr)

    return result.stdout, result.stderr


def add_files_argument(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("--files", dest="file_list", nargs='+', required=True,
                        help="Target file or directory; accepts multiple values; ")


def expand_files_argument(settings: argparse.Namespace) -> None:
    # TODO: subclass argparse.Action
    result = []
    for pattern in settings.file_list:
        result.extend(glob.glob(pattern))
    settings.file_list = result


def add_result_file_argument(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("--result-file", dest="result_file", required=True,
                        help="File for storing json results of Universum run. Set it to \"${CODE_REPORT_FILE}\" "
                             "for running from Universum, variable will be handled during run. If you run this "
                             "script separately from Universum, just name the result file or leave it empty.")


def add_python_version_argument(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("--python-version", dest="version", default="3",
                        help="Version of the python interpreter, such as 2, 3 or 3.7. "
                             "Pylint analyzer uses this parameter to select python binary for launching pylint. "
                             "For example, if the version is 3.7, it uses the following command: "
                             "'python3.7 -m pylint <...>'")


def report_to_file(issues: List[ReportData], json_file: str = None) -> None:
    issues_json = json.dumps(issues, indent=4)
    if json_file:
        with open(json_file, "w", encoding="utf-8") as f:
            f.write(issues_json)
    else:
        sys.stdout.write(issues_json)


def normalize(file: str) -> pathlib.Path:
    file_path = pathlib.Path(file)
    return file_path if file_path.is_absolute() else pathlib.Path.cwd().joinpath(file_path)



import os

from .base_output import BaseOutput


__all__ = [
    "HtmlOutput"
]


class HtmlOutput(BaseOutput):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._filename = None
        self.artifact_dir_ready = False
        self._log_buffer = []
        self._block_level = 0

    def set_artifact_dir(self, artifact_dir):
        self._filename = os.path.join(artifact_dir, "log.html")

    def open_block(self, num_str, name):
        opening_html = f'<input type="checkbox" id="{num_str}" class="hide"/>' + \
            f'<label for="{num_str}"><span class="sectionLbl">'
        closing_html = "</span></label><div>"
        name_html = f'<span class="sectionTitle">{name}</span>'
        self._log_line(f"{opening_html}{num_str} {name_html}{closing_html}", with_line_separator=False)
        self._block_level += 1

    def close_block(self, num_str, name, status):
        self._block_level -= 1
        indent = "  " * self._block_level
        closing_html = '</div><span class="nl"></span>'
        status_html = f'<span class="section{status}Status">[{status}]</span>'
        self._log_line(f"{indent} \u2514 {status_html}{closing_html}", with_line_separator=False)
        self._log_line("")

    def report_error(self, description):
        pass

    def report_skipped(self, message):
        self._log_line(message)

    def report_step(self, message, status):
        self.log(message)

    def change_status(self, message):
        pass

    def log_exception(self, line):
        self._log_line(f"Error: {line}")

    def log_stderr(self, line):
        self._log_line(f"stderr: {line}")

    def log(self, line):
        self._log_line(f"==> {line}")

    def log_external_command(self, command):
        self._log_line(f"$ {command}")

    def log_shell_output(self, line):
        self._log_line(line)

    def log_execution_start(self, title, version):
        head_content = self._build_html_head()
        html_header = f"<!DOCTYPE html><html><head>{head_content}</head><body><pre>"
        self._log_line(html_header)
        self.log(self._build_execution_start_msg(title, version))

    def log_execution_finish(self, title, version):
        self.log(self._build_execution_finish_msg(title, version))
        html_footer = "</pre></body></html>"
        self._log_line(html_footer)

    def _log_line(self, line, with_line_separator=True):
        if with_line_separator and not line.endswith(os.linesep):
            line += os.linesep
        if not self._filename:
            raise RuntimeError("Artifact directory was not set")
        if not self.artifact_dir_ready:
            self._log_buffer.append(line)
            return
        if self._log_buffer:
            self._log_and_clear_buffer()
        self._write_to_file(line)

    def _log_and_clear_buffer(self):
        for buffered_line in self._log_buffer:
            self._write_to_file(buffered_line)
        self._log_buffer = []

    def _write_to_file(self, line):
        with open(self._filename, "a", encoding="utf-8") as file:
            file.write(self._build_indent())
            file.write(line)

    def _build_indent(self):
        indent_str = []
        for x in range(0, self._block_level):
            indent_str.append("  " * x)
            indent_str.append(" |   ")
        return "".join(indent_str)

    @staticmethod
    def _build_html_head():
        css_rules = '''
            body {
                background-color: white;
                color: black;
            }

            .sectionTitle {
                color: darkslateblue;
                font-weight: bold;
            }
            .sectionSuccessStatus {
                color: green;
                font-weight: bold;
            }
            .sectionFailedStatus {
                color: red;
                font-weight: bold;
            }

            .hide {
                display: none;
            }
            .hide + label ~ div {
                display: none;
            }
            .hide + label {
                color: black;
                cursor: pointer;
                display: inline-block;
            }
            .hide:checked + label + div {
                display: block;
            }

            .hide + label + div + .nl {
                display: block;
            }
            .hide:checked + label + div + .nl::after {
                display: none;
            }

            .hide + label .sectionLbl::before {
                content: "[+] ";
            }
            .hide:checked + label .sectionLbl::before {
                content: "[-] ";
            }
        '''
        head = []
        head.append('<meta content="text/html;charset=utf-8" http-equiv="Content-Type">')
        head.append('<meta content="utf-8" http-equiv="encoding">')
        head.append(f"<style>{css_rules}</style>")
        return "".join(head)



from typing import List, Optional
import signal
import sys

from . import __version__, __title__
from .api import Api
from .github_handler import GithubHandler
from .config_creator import ConfigCreator
from .lib.ci_exception import SilentAbortException
from .lib.gravity import define_arguments_recursive, construct_component
from .lib.module_arguments import ModuleArgumentParser, ModuleNamespace, IncorrectParameterError
from .lib.utils import Uninterruptible, format_traceback
from .main import Main
from .modules.error_state import GlobalErrorState
from .nonci import Nonci
from .poll import Poll
from .submit import Submit


def define_arguments() -> ModuleArgumentParser:
    parser = ModuleArgumentParser(prog=f"python{sys.version_info.major}.{sys.version_info.minor} -m universum",
                                  description=__title__ + " " + __version__)
    parser.add_argument("--version", action="version", version=__title__ + " " + __version__)
    define_arguments_recursive(Main, parser)

    subparsers = parser.add_subparsers(title="Additional commands",
                                       metavar="{init,run,poll,submit,github-handler}",
                                       help="Use 'universum <subcommand> --help' for more info")

    def define_command(klass, command):
        command_parser = subparsers.add_parser(command)
        command_parser.set_defaults(command_parser=command_parser)
        command_parser.set_defaults(main_class=klass)
        define_arguments_recursive(klass, command_parser)

    define_command(Api, "api")
    define_command(ConfigCreator, "init")
    define_command(Poll, "poll")
    define_command(Submit, "submit")
    define_command(Nonci, "run")
    define_command(Nonci, "nonci")
    define_command(GithubHandler, "github-handler")

    return parser


def run(settings: ModuleNamespace) -> int:
    result = 0
    error_state_module = construct_component(GlobalErrorState, settings)
    main_module = construct_component(settings.main_class, settings)

    if error_state_module.is_in_error_state():
        raise IncorrectParameterError(("\n\n"+"-"*80 + "\n").join(error_state_module.get_errors()))

    main_module.out.log_execution_start(__title__, __version__)

    def signal_handler(signal_number, stack_frame):
        raise KeyboardInterrupt

    signal.signal(signal.SIGTERM, signal_handler)

    try:
        with Uninterruptible(main_module.out.log_exception) as run_function:
            run_function(main_module.execute)
            run_function(main_module.finalize)

    except SilentAbortException as e:
        result = e.application_exit_code

    except Exception as e:
        ex_traceback = sys.exc_info()[2]
        main_module.out.log_exception("Unexpected error.\n" + format_traceback(e, ex_traceback))
        main_module.out.report_build_problem("Unexpected error while executing script.")
        result = 2

    main_module.out.log_execution_finish(__title__, __version__)
    return result


def main(args: Optional[List[str]] = None) -> int:
    parser = define_arguments()
    settings = parser.parse_args(args)
    settings.main_class = getattr(settings, "main_class", Main)
    settings.command_parser = getattr(settings, "command_parser", parser)

    try:
        return run(settings)
    except IncorrectParameterError as e:
        settings.command_parser.print_usage(sys.stderr)
        sys.stderr.write("\nThe following errors were encountered:\n" + "-"*80+"\n")
        sys.stderr.write(str(e)+"\n")
        return 2
    except ImportError as e:
        print(e)
        return 2


if __name__ == "__main__":
    exit_code = main()
    sys.exit(exit_code)



import argparse
import json

from typing import List

from . import utils


def pylint_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Pylint analyzer")
    parser.add_argument("--rcfile", dest="rcfile", type=str, help="Specify a configuration file.")
    utils.add_python_version_argument(parser)
    return parser


@utils.sys_exit
@utils.analyzer(pylint_argument_parser())
def main(settings: argparse.Namespace) -> List[utils.ReportData]:
    cmd = [f"python{settings.version}", '-m', 'pylint', '-f', 'json']
    if settings.rcfile:
        cmd.append(f'--rcfile={settings.rcfile}')
    cmd.extend(settings.file_list)
    output, _ = utils.run_for_output(cmd)
    return pylint_output_parser(output)


def pylint_output_parser(output: str) -> List[utils.ReportData]:
    result: List[utils.ReportData] = []
    for data in json.loads(output):
        # pylint has its own escape rules for json output of "message" values.
        # it uses cgi.escape lib and escapes symbols <>&
        result.append(utils.ReportData(
            symbol=data["symbol"],
            message=data["message"].replace("&lt;", "<").replace("&gt;", ">").replace("&amp;", "&"),
            path=data["path"],
            line=int(data["line"])
        ))
    return result


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter  # see https://github.com/PyCQA/pylint/issues/259



from .base_output import BaseOutput

__all__ = [
    "TeamcityOutput"
]


def escape(message):
    return message.replace('\r', '').replace('|', '||').replace('\'', '|\'').replace('[', '|[').replace(']', '|]')


class TeamcityOutput(BaseOutput):
    def open_block(self, num_str, name):
        print(f"##teamcity[blockOpened name='{num_str} {escape(name)}']")

    def close_block(self, num_str, name, status):
        print(f"##teamcity[blockClosed name='{num_str} {escape(name)}']")

    def report_error(self, description):
        print(f"##teamcity[buildProblem description='<{escape(description)}>']")

    def report_skipped(self, message):
        lines = message.split("\n")
        for single_line in lines:
            print(f"##teamcity[message text='{escape(single_line)}' status='WARNING']")

    def change_status(self, message):
        print(f"##teamcity[buildStatus text='{escape(message)}']")

    def log_exception(self, line):
        lines = line.split("\n")
        for single_line in lines:
            print(f"##teamcity[message text='{escape(single_line)}' status='ERROR']")

    def log_stderr(self, line):
        lines = line.split("\n")
        for single_line in lines:
            print(f"##teamcity[message text='{escape(single_line)}' status='WARNING']")

    def log(self, line):
        print("==>", line)

    def log_external_command(self, command):
        print("$", command)

    def log_shell_output(self, line):
        print(line)



import argparse
import difflib
import os
import shutil
from pathlib import Path

from typing import Callable, List, Optional, Tuple

from . import utils


def uncrustify_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Uncrustify analyzer")
    parser.add_argument("--cfg-file", "-cf", dest="cfg_file",
                        help="Name of the configuration file of Uncrustify; "
                             "can also be set via 'UNCRUSTIFY_CONFIG' env. variable")
    parser.add_argument("--output-directory", "-od", dest="output_directory", default="uncrustify",
                        help="Directory to store fixed files, generated by Uncrustify "
                             "and HTML files with diff; the default value is 'uncrustify'"
                             "Has to be distinct from source directory")
    parser.add_argument("--report-html", dest="write_html", action="store_true", default=False,
                        help="(optional) Set to generate html reports for each modified file")
    return parser


@utils.sys_exit
@utils.analyzer(uncrustify_argument_parser())
def main(settings: argparse.Namespace) -> List[utils.ReportData]:
    if not shutil.which('uncrustify'):
        raise EnvironmentError("Please install uncrustify")
    if not settings.cfg_file and 'UNCRUSTIFY_CONFIG' not in os.environ:
        raise EnvironmentError("Please specify the '--cfg_file' parameter "
                               "or set an env. variable 'UNCRUSTIFY_CONFIG'")
    target_folder: Path = utils.normalize(settings.output_directory)
    if target_folder.exists() and target_folder.samefile(Path.cwd()):
        raise EnvironmentError("Target and source folders for uncrustify are not allowed to match")
    html_diff_file_writer: Optional[Callable[[Path, List[str], List[str]], None]] = None
    if settings.write_html:
        wrapcolumn, tabsize = _get_wrapcolumn_tabsize(settings.cfg_file)
        html_diff_file_writer = HtmlDiffFileWriter(target_folder, wrapcolumn, tabsize)

    files: List[Tuple[Path, Path]] = []
    for src_file in settings.file_list:
        src_file_absolute = utils.normalize(src_file)
        src_file_relative = src_file_absolute.relative_to(Path.cwd())
        target_file_absolute: Path = target_folder.joinpath(src_file_relative)
        files.append((src_file_absolute, target_file_absolute))
    cmd = ["uncrustify", "-q", "-c", settings.cfg_file, "--prefix", settings.output_directory]
    cmd.extend(settings.file_list)
    utils.run_for_output(cmd)
    return uncrustify_output_parser(files, html_diff_file_writer)


class HtmlDiffFileWriter:

    def __init__(self, target_folder: Path, wrapcolumn: int, tabsize: int) -> None:
        self.target_folder = target_folder
        self.differ = difflib.HtmlDiff(tabsize=tabsize, wrapcolumn=wrapcolumn)

    def __call__(self, file: Path, src: List[str], target: List[str]) -> None:
        file_relative = file.relative_to(Path.cwd())
        out_file_name: str = str(file_relative).replace('/', '_') + '.html'
        with open(self.target_folder.joinpath(out_file_name), 'w', encoding="utf-8") as out_file:
            out_file.write(self.differ.make_file(src, target, context=False))


def uncrustify_output_parser(files: List[Tuple[Path, Path]],
                             write_diff_file: Optional[Callable[[Path, List[str], List[str]], None]]
                             ) -> List[utils.ReportData]:
    result: List[utils.ReportData] = []
    for src_file, uncrustify_file in files:
        with open(src_file, encoding="utf-8") as src:
            src_lines = src.readlines()
        with open(uncrustify_file, encoding="utf-8") as fixed:
            fixed_lines = fixed.readlines()

        issues = _get_issues_from_diff(src_file, src_lines, fixed_lines)
        if issues and write_diff_file:
            write_diff_file(src_file, src_lines, fixed_lines)
        result.extend(issues)
    return result


def _get_wrapcolumn_tabsize(cfg_file: str) -> Tuple[int, int]:
    with open(cfg_file, encoding="utf-8") as config:
        for line in config.readlines():
            if line.startswith("code_width"):
                wrapcolumn = int(line.split()[2])
            if line.startswith("input_tab_size"):
                tabsize = int(line.split()[2])
    return wrapcolumn, tabsize


def _get_issues_from_diff(src_file: Path, src: List[str], target: List[str]) -> List[utils.ReportData]:
    result = []
    matching_blocks: List[difflib.Match] = \
        difflib.SequenceMatcher(a=src, b=target).get_matching_blocks()
    previous_match = matching_blocks[0]
    for match in matching_blocks[1:]:
        block = _get_mismatching_block(previous_match, match, src, target)
        previous_match = match
        if not block:
            continue
        line, before, after = block
        path: Path = src_file.relative_to(Path.cwd())
        message = _get_issue_message(before, after)
        result.append(utils.ReportData(
            symbol="Uncrustify Code Style issue",
            message=message,
            path=str(path),
            line=line
        ))

    return result


def _get_issue_message(before: str, after: str) -> str:
    # The maximum number of lines to write separate comments for
    # If exceeded, summarized comment will be provided instead
    max_lines = 11
    diff_size = len(before.splitlines())
    if diff_size > max_lines:
        message = f"\nLarge block of code ({diff_size} lines) has issues\n" + \
                  f"Non-compliant code blocks exceeding {max_lines} lines are not reported\n"
    else:
        # Message with before&after
        message = f"\nOriginal code:\n```diff\n{before}```\n" + \
                  f"Uncrustify generated code:\n```diff\n{after}```\n"
    return message


def _get_mismatching_block(previous_match: difflib.Match,  # src[a:a+size] = target[b:b+size]
                           match: difflib.Match,
                           src: List[str], target: List[str]
                           ) -> Optional[Tuple[int, str, str]]:
    previous_match_end_in_src = previous_match.a + previous_match.size
    previous_match_end_in_target = previous_match.b + previous_match.size
    match_start_in_src = match.a
    match_start_in_target = match.b
    if previous_match_end_in_src == match_start_in_src:
        return None
    line = match_start_in_src
    before = _get_text_for_block(previous_match_end_in_src - 1, match_start_in_src, src)
    after = _get_text_for_block(previous_match_end_in_target - 1, match_start_in_target, target)
    return line, before, after


def _get_text_for_block(start: int, end: int, lines: List[str]) -> str:
    return _replace_invisible_symbols(''.join(lines[start: end]))


def _replace_invisible_symbols(line: str) -> str:
    for old_str, new_str in zip([" ", "\t", "\n"], ["\u00b7", "\u2192\u2192\u2192\u2192", "\u2193\u000a"]):
        line = line.replace(old_str, new_str)
    return line


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter  # see https://github.com/PyCQA/pylint/issues/259



from setuptools import setup, find_packages

import universum


def readme():
    with open('README.md', encoding="utf-8") as f:
        return f.read()


p4 = ('pip>=19', 'p4python>=2019.1')

git = 'gitpython>=3.0.5'

github = (git, 'cryptography', 'pygithub')

vcs = p4 + github

docs = ('sphinx==3.4.3', 'sphinx-argparse', 'sphinx_rtd_theme')  # This extra is required for RTD to generate documentation

setup(
    name=universum.__title__,
    version=universum.__version__,
    description='Unifier of Continuous Integration',
    long_description=readme(),
    long_description_content_type='text/markdown',
    author='Ivan Keliukh <i.keliukh@samsung.com>, Kateryna Dovgan <k.dovgan@samsung.com>',
    license='BSD',
    packages=find_packages(exclude=['tests', 'tests.*']),
    py_modules=['universum'],
    python_requires='>=3.6',
    install_requires=[
        'glob2',
        'requests',
        'sh',
        'lxml',
        'typing-extensions'
    ],
    extras_require={
        'p4': [p4],
        'git': [git],
        'github': [github],
        'docs': [docs],
        'test': [
            vcs,
            docs,
            'docker',
            'httpretty',
            'mock',
            'pytest',
            'pylint',
            'pytest-pylint',
            'teamcity-messages',
            'pytest-cov',
            'coverage',
            'mypy',
            'types-requests',
            'selenium'
        ]
    }
)


if __name__ == "__main__":
    print("Please use 'sudo pip install .' instead of launching this script")



from .base_server import BaseServerForHostingBuild, BaseServerForTrigger

__all__ = [
    "LocalServer"
]


class LocalServer(BaseServerForHostingBuild, BaseServerForTrigger):

    def report_build_location(self) -> str:
        return "Local build. No external logs provided"

    def artifact_path(self, local_artifacts_dir: str, item: str) -> str:
        return local_artifacts_dir + "/" + item
